#include <windows.h>
#include <stdio.h>
#include <math.h>

inline float Clamp(float f, float low, float high)
{
	if (f < low)
		f = low;
	if (f > high)
		f = high;
	return f;
}

inline int Wrap(int coord, int boundary)
{
	if (coord < 0)
		coord = boundary + coord;
	else if (coord >= boundary)
		coord -= boundary;

	return coord;
}

class CImage
{
public:
	int m_iWidth, m_iHeight, m_iBPP;
	unsigned char* m_pImage;

	CImage(void)
	{
		m_pImage = NULL;
		m_iWidth = m_iHeight = m_iBPP = 0;
	}
	~CImage(void)
	{
		if (m_pImage)
			delete[] m_pImage;
	}

	bool LoadBMP(char* szFile)
	{
		BITMAPFILEHEADER bmfh;
		BITMAPINFOHEADER bmih;
		BITMAPINFO* pbmi;
		RGBQUAD* pPalette;
		unsigned char* pRaster;

		FILE* pFile = fopen(szFile , "rb"); // open the file
		if (!pFile)
			return false;

		// load the headers
		fread((char*)&bmfh, sizeof(BITMAPFILEHEADER), 1, pFile);
		fread((char*)&bmih, sizeof(BITMAPINFOHEADER), 1, pFile);

		if (bmfh.bfType != 'MB') // check that it's actually a bitmap
			return false;

		// get the vital statistics
		m_iBPP = bmih.biBitCount;
		m_iWidth = bmih.biWidth;
		m_iHeight = (bmih.biHeight > 0) ? bmih.biHeight : -bmih.biHeight; // absolute height - negative height = flipped image
		int iBytesPerRow = m_iWidth * m_iBPP / 8;
		iBytesPerRow += (4 - iBytesPerRow % 4) % 4; // 4-byte alignment

		// if it's not 24-bit, load the palette
		if (m_iBPP == 24)
			pbmi = (BITMAPINFO*)malloc(sizeof(BITMAPINFO));
		else {
			pbmi = (BITMAPINFO*)malloc(sizeof(BITMAPINFOHEADER)+(1<<m_iBPP)*sizeof(RGBQUAD));
			pPalette = (RGBQUAD*)((char*)pbmi + sizeof(BITMAPINFOHEADER));
			fread((char*)pPalette, sizeof(RGBQUAD) * (1<<m_iBPP), 1, pFile);
		}
		pbmi->bmiHeader = bmih;

		// load Raster
		fseek(pFile, bmfh.bfOffBits, SEEK_SET);

		pRaster = new unsigned char[iBytesPerRow*m_iHeight];

		// (if height is positive the bmp is bottom-up, read it reversed)
		if (bmih.biHeight > 0)
			for (int n = (m_iHeight-1); n >= 0; n--)
				fread(pRaster + iBytesPerRow*n, 1, iBytesPerRow, pFile);
		else
			fread(pRaster, 1, iBytesPerRow*m_iHeight, pFile);

		fclose(pFile);

		// next, make a 24-bit image of it
		m_pImage = new unsigned char[m_iWidth * m_iHeight * 3];
		if (m_iBPP == 24)
			memcpy(m_pImage, pRaster, m_iWidth * m_iHeight * 3); // straight copy
		else {
			// 8-bit image - paletted
			for (int y = 0; y < m_iHeight; y++) {
				for (int x = 0; x < m_iWidth; x++) {
					m_pImage[(x + y*m_iWidth)*3] = pPalette[pRaster[x + y*m_iWidth]].rgbRed;
					m_pImage[(x + y*m_iWidth)*3 + 1] = pPalette[pRaster[x + y*m_iWidth]].rgbGreen;
					m_pImage[(x + y*m_iWidth)*3 + 2] = pPalette[pRaster[x + y*m_iWidth]].rgbBlue;
				}
			}
		}

		if (m_iBPP == 24) {
			// 24-bit BMPs are stored BGR - make it RGB
			for (int i = 0; i < (m_iWidth*m_iHeight*3); i += 3) {
				unsigned char temp = m_pImage[i];
				m_pImage[i] = m_pImage[i+2];
				m_pImage[i+2] = temp;
			}
		}
		
		m_iBPP = 24; // if it was 8 BPP when loaded, it's 24 BPP now

		free(pbmi);
		delete[] pRaster;

		return true; // success
	}

	bool WriteBMP(char* szFile)
	{
		FILE* pFile = fopen(szFile, "wb");
		if (!pFile)
			return false;

		//fwrite(m_pImage, 1, m_iWidth*m_iHeight*3, pFile);

		BITMAPFILEHEADER bmfh;
		BITMAPINFOHEADER bmih;

		bmfh.bfType = 'MB';
		bmfh.bfOffBits = sizeof(BITMAPINFOHEADER) + sizeof(BITMAPFILEHEADER);
		bmfh.bfSize = m_iWidth*m_iHeight*3;

		bmih.biSize = sizeof(BITMAPINFOHEADER);
		bmih.biWidth = m_iWidth;
		bmih.biHeight = m_iHeight;
		bmih.biPlanes = 1;
		bmih.biBitCount = 24;
		bmih.biCompression = BI_RGB;
		bmih.biSizeImage = 0;
		bmih.biXPelsPerMeter = 1;
		bmih.biYPelsPerMeter = 1;
		bmih.biClrUsed = 0;
		bmih.biClrImportant = 0;

		fwrite(&bmfh, sizeof(BITMAPFILEHEADER), 1, pFile);
		fwrite(&bmih, sizeof(BITMAPINFOHEADER), 1, pFile);

		// bitmaps are stored upside-down
		for (int y = (m_iHeight-1); y >= 0; y--)
			fwrite(m_pImage + y*m_iWidth*3, m_iWidth*3, 1, pFile);

		fclose(pFile);

		return true;
	}
};

int main(int argc, char* argv[])
{
	CImage img, imgOut;

	if (img.LoadBMP(argv[1]))
	{
		imgOut.m_pImage = new unsigned char[img.m_iWidth*img.m_iHeight*3];
		imgOut.m_iBPP = 24;
		imgOut.m_iWidth = img.m_iWidth;
		imgOut.m_iHeight = img.m_iHeight;

		float nx, ny, nz, nmag;
		float changeNeg, changePos;
		float h1, h2, h3;

		for (int y = 0; y < img.m_iHeight; y++)
		{
			for (int x = 0; x < img.m_iWidth; x++)
			{
				// CHANGE IN X

				h1 = (float)img.m_pImage[(Wrap(x-1, img.m_iWidth) + y*img.m_iWidth)*3];
				h2 = (float)img.m_pImage[(x + y*img.m_iWidth)*3];
				h3 = (float)img.m_pImage[(Wrap(x+1, img.m_iWidth) + y*img.m_iWidth)*3];

				changeNeg = (h2 - h1);
				changePos = (h3 - h2);
				
				nx = -(changePos + changeNeg) / 2;

				// CHANGE IN Y

				h1 = (float)img.m_pImage[(x + Wrap(y-1, img.m_iHeight)*img.m_iWidth)*3];
				h2 = (float)img.m_pImage[(x + y*img.m_iWidth)*3];
				h3 = (float)img.m_pImage[(x + Wrap(y+1, img.m_iHeight)*img.m_iWidth)*3];

				changeNeg = (h2 - h1);
				changePos = (h3 - h2);
				
				ny = (changePos + changeNeg) / 2;

				// SCALE OF BUMPINESS

				nz = (float)atof(argv[2]);

				// NORMALIZE

				nmag = sqrtf(nx*nx + ny*ny + nz*nz);
				nx /= nmag;
				ny /= nmag;
				nz /= nmag;

				// PUT INTO NEW IMAGE

				imgOut.m_pImage[(x + y*img.m_iWidth)*3 + 2] = (unsigned char)Clamp(((nx + 1) / 2) * 255, 0, 255);
				imgOut.m_pImage[(x + y*img.m_iWidth)*3 + 1] = (unsigned char)Clamp(((ny + 1) / 2) * 255, 0, 255);
				imgOut.m_pImage[(x + y*img.m_iWidth)*3 + 0] = (unsigned char)Clamp(((nz + 1) / 2) * 255, 0, 255);
			}
		}

		imgOut.WriteBMP(argv[3]);
	}

	return 0;
}